from typing import Any, Callable, get_args

from pydantic import BaseModel, ValidationError
from starlette.datastructures import FormData

from . import tags
from .requests import Request

try:
    from typing import Self  # type: ignore [attr-defined]
except ImportError:
    # NOTE: Remove once Python 3.10 support is dropped
    Self = "AirForm"  # type: ignore [assignment]


class AirForm:
    """
    A form handler that validates incoming form data against a Pydantic model.
    To be used with FastAPI's dependency injection system.

        class CheeseModel(pydantic.BaseModel):
            name: str
            age: int

        class CheeseForm(air.AirForm):
            model = CheeseModel

        @app.post("/cheese")
        async def cheese_form(cheese: Annotated[CheeseForm, Depends(CheeseForm())]):
            if cheese.is_valid:
                return air.Html(air.H1(cheese.data.name))
            return air.Html(air.H1(air.RawHTML(str(len(cheese.errors)))))

    NOTE: This is named AirForm to avoid collisions with tags.Form
    """

    model: type[BaseModel] | None = None
    data: Any = None  # TODO change type to something more specific
    initial_data = None
    errors = None
    is_valid = None

    def __init__(self, initial_data: dict | None = None):
        if self.model is None:
            raise NotImplementedError("model")
        self.initial_data = initial_data

    async def __call__(self, form_data: dict[Any, Any] | FormData) -> Self:  # ty: ignore [invalid-type-form]
        self.validate(form_data)
        return self

    def validate(self, form_data: dict[Any, Any] | FormData):
        if self.model is None:
            raise NotImplementedError("model")
        try:
            self.data = self.model(**form_data)
            self.is_valid = True
        except ValidationError as e:
            self.errors = e.errors()
            self.is_valid = False

    @classmethod
    async def from_request(cls, request: Request) -> Self:  # ty:ignore [invalid-type-form]
        form_data = await request.form()
        self = cls()
        await self(form_data=form_data)
        return self

    @property
    def widget(self) -> Callable:
        """Widget for rendering of form in HTML

        If you want a custom widget, replace with a function that accepts:

            - model: BaseModel
            - data: dict|None
            - errors:dict|None=None
        """
        return default_form_widget

    def render(self) -> tags.SafeStr:
        return tags.SafeStr(
            self.widget(model=self.model, data=self.initial_data, errors=self.errors)
        )


def pydantic_type_to_html_type(field_info: Any) -> str:
    """Return HTML type from pydantic type.

    Default to 'text' for unknown types.
    """
    special_fields = [
        "hidden",
        "email",
        "password",
        "url",
        "datedatetime-local",
        "month",
        "time",
        "color",
        "file",
    ]
    for field in special_fields:
        if field_info.json_schema_extra and field_info.json_schema_extra.get(
            field, False
        ):
            return field

    return {int: "number", float: "number", bool: "checkbox", str: "text"}.get(
        field_info.annotation, "text"
    )


def errors_to_dict(errors: list[dict] | None) -> dict[str, dict]:
    "Converts a pydantic error list to a dictionary for easier reference."
    if errors is None:
        return {}
    return {error["loc"][0]: error for error in errors}


def default_form_widget(
    model: type[BaseModel], data: dict | None = None, errors: list | None = None
) -> str:
    error_dict = errors_to_dict(errors)
    fields = []
    for field_name, field_info in model.model_fields.items():
        field_type = field_info.annotation

        # Handle optional types (Union with None)
        if (
            hasattr(field_type, "__origin__")
            and field_type.__origin__ is type(None | str).__origin__  # type: ignore
        ):
            # This is a Union type, get the non-None type
            args = get_args(field_type)
            field_type = next((arg for arg in args if arg is not type(None)), str)
        input_type = pydantic_type_to_html_type(field_info)
        kwargs = {}
        # Inject values
        if data is not None and field_name in data:
            kwargs["value"] = data[field_name]

        if error := error_dict.get(field_name, False):
            kwargs["aria-invalid"] = "true"
        fields.append(
            tags.Label(
                field_name,
                tags.Input(name=field_name, type=input_type, id=field_name, **kwargs),
                tags.Small("Please correct this error.", id=f"{field_name}-error")
                if error
                else "",
            )
        )

    return tags.Fieldset(*fields).render()



from .tags import Base, Body, Head, Html, Link, Main, Meta, Script, Style, Tag, Title

# ruff: noqa F841
HEAD_TAG_TYPES: tuple[Tag] = (Title, Style, Meta, Link, Script, Base)  # type: ignore [assignment]


def filter_body_tags(tags) -> list:
    """Given a list of tags, only list the ones that belong in body."""
    return [t for t in tags if not isinstance(t, HEAD_TAG_TYPES)]  # type: ignore [arg-type]


def filter_head_tags(tags) -> list:
    """Given a list of tags, only list the ones that belong in head."""
    return [t for t in tags if isinstance(t, HEAD_TAG_TYPES)]  # type: ignore [arg-type]


def picocss(*children, htmx: bool = True, **kwargs):
    """Renders the basic layout with PicoCSS and HTMX for quick prototyping

    1. At the top level HTML head tags are put in the `<head>` tag
    2. Otherwise everything is put in the `<body>`
    3. HTMX is the default, change with the `htmx` keyword argument

    Note: `PicoCSS` is a quick prototyping tool. It isn't designed to be extensible.
        Rather the `pico` layout function makes it easy to roll out quick demonstrations and proofs-of-concept.
        For more advanced layouts like Eidos or a full-fledged PicoCSS implementation,
        you'll have to create your own layouts.
    """
    body_tags = filter_body_tags(children)  # type: ignore [arg-type]
    head_tags = filter_head_tags(children)  # type: ignore [arg-type]
    if htmx:
        head_tags.insert(
            0,
            Script(
                src="https://cdn.jsdelivr.net/npm/htmx.org@2.0.6/dist/htmx.min.js",
                integrity="sha384-Akqfrbj/HpNVo8k11SXBb6TlBWmXXlYQrCSqEWmyKJe+hDm3Z/B2WVG4smwBkRVm",
                crossorigin="anonymous",
            ),
        )
    return Html(
        Head(
            Link(
                rel="stylesheet",
                href="https://cdn.jsdelivr.net/npm/@picocss/pico@2/css/pico.min.css",
            ),
            *head_tags,
        ),
        Body(Main(*body_tags, class_="container")),
    ).render()


def mvpcss(*children, htmx: bool = True, **kwargs):
    """Renders the basic layout with MVP.css and HTMX for quick prototyping

    1. At the top level HTML head tags are put in the `<head>` tag
    2. Otherwise everything is put in the `<body>`
    3. HTMX is the default, change with the `htmx` keyword argument

    Note: `MVP.css` is a quick prototyping tool. It isn't designed to be extensible.
        Rather the `mvpcss` layout function makes it easy to roll out quick demonstrations and proofs-of-concept.
        For more advanced layouts like Eidos or a full-fledged PicoCSS implementation,
        you'll have to create your own layouts.
    """
    body_tags = filter_body_tags(children)  # type: ignore [arg-type]
    head_tags = filter_head_tags(children)  # type: ignore [arg-type]
    if htmx:
        head_tags.insert(
            0,
            Script(
                src="https://cdn.jsdelivr.net/npm/htmx.org@2.0.6/dist/htmx.min.js",
                integrity="sha384-Akqfrbj/HpNVo8k11SXBb6TlBWmXXlYQrCSqEWmyKJe+hDm3Z/B2WVG4smwBkRVm",
                crossorigin="anonymous",
            ),
        )
    return Html(
        Head(
            Link(rel="stylesheet", href="https://unpkg.com/mvp.css"),
            *head_tags,
        ),
        Body(Main(*body_tags)),
    ).render()



"""Adds s-expression HTML tags to air."""

import html
from functools import cached_property
from xml.etree import ElementTree as ET


def html_to_airtags(html_text: str, air_prefix: bool = True) -> str:
    def convert_attrs(attrs):
        parts = []
        for key, value in attrs.items():
            if key in {"class", "for", "id_"}:
                key += "_"
            parts.append(f"{key}='{html.escape(value, quote=True)}'")
        return parts

    def convert_node(el, indent=0, air_prefix: bool = True):
        ind = "    " * indent
        if air_prefix:
            tag = f"air.{el.tag.capitalize()}"
        else:
            tag = el.tag.capitalize()

        args = []
        # Add text before children if any
        if el.text and el.text.strip():
            args.append(repr(el.text.strip()))

        # Add children recursively
        for child in el:
            args.append(convert_node(child, indent + 1, air_prefix=air_prefix))
            if child.tail and child.tail.strip():
                args.append(repr(child.tail.strip()))

        # Add attributes
        attr_args = convert_attrs(el.attrib)
        all_args = args + attr_args

        if all_args:
            if len(all_args) == 1:
                return f"{ind}{tag}(\n{all_args[0]})\n"
            else:
                joined = ",\n".join("    " * (indent + 1) + arg for arg in all_args)
                return f"{ind}{tag}(\n{joined}\n{ind})"
        else:
            return f"{ind}{tag}()"

    root = ET.fromstring(html_text)
    return convert_node(root, air_prefix=air_prefix)


def clean_html_attr_key(key: str) -> str:
    """Clean up HTML attribute keys to match the standard W3C HTML spec.

    Args:
        key: An uncleaned HTML attribute key

    Returns:

        Cleaned HTML attribute key
    """
    # If a "_"-suffixed proxy for "class", "for", or "id" is used,
    # convert it to its normal HTML equivalent.
    key = dict(class_="class", for_="for", id_="id").get(key, key)
    # Remove leading underscores and replace underscores with dashes
    return key.lstrip("_").replace("_", "-")


class Tag:
    self_closing = False

    def __init__(self, *children, **kwargs):
        """Sets four attributes, name, module, children, and attrs.
        These are important for Starlette view responses, as nested objects
        get auto-serialized to JSON and need to be rebuilt. Without
        the values of these attributes, the object reconstruction can't occur"""
        self._name = self.__class__.__name__
        self._module = self.__class__.__module__
        self._children, self._attrs = children, kwargs

    @property
    def name(self) -> str:
        return self._name.lower()

    @property
    def attrs(self) -> str:
        if not self._attrs:
            return ""
        attrs = []
        for k, v in self._attrs.items():
            if isinstance(v, bool) and v is True:
                # Add single word attribute like "selected"
                attrs.append(k)
            elif isinstance(v, bool) and v is False:
                # Skip single word attribute like "selected"
                continue
            else:
                attrs.append(f'{clean_html_attr_key(k)}="{v}"')
        return " " + " ".join(attrs)

    @cached_property
    def children(self):
        elements = []
        for child in self._children:
            if isinstance(child, Tag):
                elements.append(child.render())
            elif isinstance(child, SafeStr):
                elements.append(child)
            elif isinstance(child, str):
                elements.append(html.escape(child))
            else:
                # If the type isn't supported, we just convert to `str`
                # and then escape it for safety. This matches to what most
                # template tools do, which prevents hard bugs in production
                # from stopping users cold.
                elements.append(html.escape(str(child)))
        return "".join(elements)

    def render(self) -> str:
        if self.self_closing:
            return f"<{self.name}{self.attrs} />"
        return f"<{self.name}{self.attrs}>{self.children}</{self.name}>"


class CaseTag(Tag):
    """This is for case-sensitive tags like those used in SVG generation."""

    @property
    def name(self) -> str:
        return self._name[0].lower() + self._name[1:]


# Utilities


class SafeStr(str):
    """A string subclass that doesn't trigger html.escape() when called by Tag.render()

    Example:
        sample = SafeStr('Hello, world')
    """

    def __new__(cls, value):
        obj = super().__new__(cls, value)
        return obj

    def __repr__(self):
        return super().__repr__()


# Special tags


class Html(Tag):
    """Defines the root of an HTML document"""

    def render(self) -> str:
        return f"""<!doctype html><html{self.attrs}>{self.children}</html>"""


class RawHTML(Tag):
    """Renders raw HTML content without escaping.

    Args:
        html_string: A single string containing raw HTML to render

    Raises:
        TypeError: If non-string content is provided
        ValueError: If multiple arguments are provided

    Example:
        >>> RawHTML('<strong>Bold</strong> text')
        '<strong>Bold</strong> text'

        >>> # Use with other tags
        >>> Div(
        ...     P("Safe content"),
        ...     RawHTML('<hr class="divider">'),
        ...     P("More safe content")
        ... )
    """

    def __init__(self, *args, **kwargs):
        """Initialize RawHTML with a single string argument.

        Args:
            *args: Should be exactly one string argument
            **kwargs: Ignored (for consistency with Tag interface)
        """
        if len(args) > 1:
            raise ValueError("RawHTML accepts only one string argument")

        html_string = args[0] if args else ""

        if not isinstance(html_string, str):
            raise TypeError("RawHTML only accepts string content")

        super().__init__(html_string)

    def render(self) -> str:
        """Render the raw HTML string without escaping."""
        return self._children[0] if self._children else ""


class NoEscapeTag(Tag):
    """Custom tag that does not escape its children.

    This is used for tags like Script and Style where content
    should not be HTML-escaped.
    """

    def render(self) -> str:
        """Render the tag with unescaped content."""
        content = self._children[0] if self._children else ""
        return f"<{self.name}{self.attrs}>{content}</{self.name}>"


class Script(NoEscapeTag):
    """Defines a client-side script

    Warning: Script tag does not protect against code injection.
    """


class Style(NoEscapeTag):
    """Defines style information for a document

    Warning: Style tag does not protect against code injection.
    """


# HTML tag attribute map

html_attributes = {
    "A": [
        "href",
        "target",
        "download",
        "rel",
        "hreflang",
        "typereferrerpolicy",
        "media",
        "ping",
        "class_",
        "id",
    ]
}


def locals_cleanup(local_data, obj):
    """Converts arguments to kwargs per the html_attributes structure"""
    data = {}
    for attr in html_attributes.get(obj.__class__.__name__, []):
        if local_data.get(attr) is not None:
            data[attr] = local_data[attr]
    return data


# Stock tags


class A(Tag):
    """Defines a hyperlink"""

    def __init__(
        self,
        *children,
        href: str | None = None,
        target: str | None = None,
        download: str | None = None,
        rel: str | None = None,
        hreflang: str | None = None,
        type: str | None = None,
        referrerpolicy: str | None = None,
        media: str | None = None,
        ping: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        **kwargs,
    ):
        super().__init__(*children, **kwargs | locals_cleanup(locals(), self))


class Abbr(Tag):
    """Defines an abbreviation or an acronym"""

    pass


class Address(Tag):
    """Defines contact information for the author/owner of a document"""

    pass


class Area(Tag):
    """Defines an area inside an image map"""

    self_closing = True


class Article(Tag):
    """Defines an article"""

    pass


class Aside(Tag):
    """Defines content aside from the page content"""

    pass


class Audio(Tag):
    """Defines embedded sound content"""

    pass


class B(Tag):
    """Defines bold text"""

    pass


class Base(Tag):
    """Specifies the base URL/target for all relative URLs in a document"""

    self_closing = True


class Bdi(Tag):
    """Isolates a part of text that might be formatted in a different direction from other text outside it"""

    pass


class Bdo(Tag):
    """Overrides the current text direction"""

    pass


class Blockquote(Tag):
    """Defines a section that is quoted from another source"""

    pass


class Body(Tag):
    """Defines the document's body"""

    pass


class Br(Tag):
    """Defines a single line break"""

    self_closing = True


class Button(Tag):
    """Defines a clickable button"""

    pass


class Canvas(Tag):
    """Used to draw graphics, on the fly, via scripting (usually JavaScript)"""

    pass


class Caption(Tag):
    """Defines a table caption"""

    pass


class Cite(Tag):
    """Defines the title of a work"""

    pass


class Code(Tag):
    """Defines a piece of computer code"""

    pass


class Col(Tag):
    """Specifies column properties for each column within a <colgroup> element"""

    self_closing = True


class Colgroup(Tag):
    """Specifies a group of one or more columns in a table for formatting"""

    pass


class Data(Tag):
    """Adds a machine-readable translation of a given content"""

    pass


class Datalist(Tag):
    """Specifies a list of pre-defined options for input controls"""

    pass


class Dd(Tag):
    """Defines a description/value of a term in a description list"""

    pass


class Del(Tag):
    """Defines text that has been deleted from a document"""

    pass


class Details(Tag):
    """Defines additional details that the user can view or hide"""

    pass


class Dfn(Tag):
    """Specifies a term that is going to be defined within the content"""

    pass


class Dialog(Tag):
    """Defines a dialog box or window"""

    pass


class Div(Tag):
    """Defines a section in a document"""

    pass


class Dl(Tag):
    """Defines a description list"""

    pass


class Dt(Tag):
    """Defines a term/name in a description list"""

    pass


class Em(Tag):
    """Defines emphasized text"""

    pass


class Embed(Tag):
    """Defines a container for an external application"""

    self_closing = True


class Fieldset(Tag):
    """Groups related elements in a form"""

    pass


class Figcaption(Tag):
    """Defines a caption for a <figure> element"""

    pass


class Figure(Tag):
    """Specifies self-contained content"""

    pass


class Footer(Tag):
    """Defines a footer for a document or section"""

    pass


class Form(Tag):
    """Defines an HTML form for user input"""

    pass


class H1(Tag):
    """H1 header"""

    pass


class H2(Tag):
    """H2 header"""

    pass


class H3(Tag):
    """H3 header"""

    pass


class H4(Tag):
    """H4 header"""

    pass


class H5(Tag):
    """H5 header"""

    pass


class H6(Tag):
    """H6 header"""

    pass


class Head(Tag):
    """Contains metadata/information for the document"""

    pass


class Header(Tag):
    """Defines a header for a document or section"""

    pass


class Hgroup(Tag):
    """Defines a header and related content"""

    pass


class Hr(Tag):
    """Defines a thematic change in the content"""

    self_closing = True


class I(Tag):  # noqa: E742
    """Defines a part of text in an alternate voice or mood"""

    pass


class Iframe(Tag):
    """Defines an inline frame"""

    pass


class Img(Tag):
    """Defines an image"""

    self_closing = True


class Input(Tag):
    """Defines an input control"""

    self_closing = True


class Ins(Tag):
    """Defines a text that has been inserted into a document"""

    pass


class Kbd(Tag):
    """Defines keyboard input"""

    pass


class Label(Tag):
    """Defines a label for an <input> element"""

    pass


class Legend(Tag):
    """Defines a caption for a <fieldset> element"""

    pass


class Li(Tag):
    """Defines a list item"""

    pass


class Link(Tag):
    """Defines the relationship between a document and an external resource (most used to link to style sheets)"""

    self_closing = True


class Main(Tag):
    """Specifies the main content of a document"""

    pass


class Map(Tag):
    """Defines an image map"""

    pass


class Mark(Tag):
    """Defines marked/highlighted text"""

    pass


class Menu(Tag):
    """Defines an unordered list"""

    pass


class Meta(Tag):
    """Defines metadata about an HTML document"""

    self_closing = True


class Meter(Tag):
    """Defines a scalar measurement within a known range (a gauge)"""

    pass


class Nav(Tag):
    """Defines navigation links"""

    pass


class Noscript(Tag):
    """Defines an alternate content for users that do not support client-side scripts"""

    pass


class Object(Tag):
    """Defines a container for an external application"""

    pass


class Ol(Tag):
    """Defines an ordered list"""

    pass


class Optgroup(Tag):
    """Defines a group of related options in a drop-down list"""

    pass


class Option(Tag):
    """Defines an option in a drop-down list"""

    pass


class Output(Tag):
    """Defines the result of a calculation"""

    pass


class P(Tag):
    """Defines a paragraph"""

    pass


class Param(Tag):
    """Defines a parameter for an object"""

    self_closing = True


class Picture(Tag):
    """Defines a container for multiple image resources"""

    pass


class Pre(Tag):
    """Defines preformatted text"""

    pass


class Progress(Tag):
    """Represents the progress of a task"""

    pass


class Q(Tag):
    """Defines a short quotation"""

    pass


class Rp(Tag):
    """Defines what to show in browsers that do not support ruby annotations"""

    pass


class Rt(Tag):
    """Defines an explanation/pronunciation of characters (for East Asian typography)"""

    pass


class Ruby(Tag):
    """Defines a ruby annotation (for East Asian typography)"""

    pass


class S(Tag):
    """Defines text that is no longer correct"""

    pass


class Samp(Tag):
    """Defines sample output from a computer program"""

    pass


class Search(Tag):
    """Defines a search section"""

    pass


class Section(Tag):
    """Defines a section in a document"""

    pass


class Select(Tag):
    """Defines a drop-down list"""

    pass


class Small(Tag):
    """Defines smaller text"""

    pass


class Source(Tag):
    """Defines multiple media resources for media elements (<video> and <audio>)"""

    pass


class Span(Tag):
    """Defines a section in a document"""

    pass


class Strong(Tag):
    """Defines important text"""

    pass


class Sub(Tag):
    """Defines subscripted text"""

    pass


class Summary(Tag):
    """Defines a visible heading for a <details> element"""

    pass


class Sup(Tag):
    """Defines superscripted text"""

    pass


class Table(Tag):
    """Defines a table"""

    pass


class Tbody(Tag):
    """Groups the body content in a table"""

    pass


class Td(Tag):
    """Defines a cell in a table"""

    pass


class Template(Tag):
    """Defines a container for content that should be hidden when the page loads"""

    pass


class Textarea(Tag):
    """Defines a multiline input control (text area)"""

    pass


class Tfoot(Tag):
    """Groups the footer content in a table"""

    pass


class Th(Tag):
    """Defines a header cell in a table"""

    pass


class Thead(Tag):
    """Groups the header content in a table"""

    pass


class Time(Tag):
    """Defines a specific time (or datetime)"""

    pass


class Title(Tag):
    """Defines a title for the document"""

    pass


class Tr(Tag):
    """Defines a row in a table"""

    pass


class Track(Tag):
    """Defines text tracks for media elements (<video> and <audio>)"""

    self_closing = True


class U(Tag):
    """Defines some text that is unarticulated and styled differently from normal text"""

    pass


class Ul(Tag):
    """Defines an unordered list"""

    pass


class Var(Tag):
    """Defines a variable"""

    pass


class Video(Tag):
    """Defines embedded video content"""

    pass


class Wbr(Tag):
    """Defines a possible line-break"""

    self_closing = True



import air.svg as sg


def test_atag_no_attrs_no_children():
    assert sg.A().render() == "<a></a>"


def test_cased_tag_no_children():
    assert sg.AnimateMotion().render() == "<animateMotion></animateMotion>"


def test_cased_tags_with_children():
    assert (
        sg.AnimateMotion(sg.ClipPath()).render()
        == "<animateMotion><clipPath></clipPath></animateMotion>"
    )



from fastapi import Header
from starlette.requests import Request as Request  # noqa: F401


def is_htmx_request(hx_request: str = Header(default=None)) -> bool:
    """Dependency injection function used to determine if request is coming from HTMX. Checks if 'hx-request' header is in the HTTP request.

    Returns:
        bool: Whether or not a request is coming from an HTMX action.

    Example:

        import air

        app = air.App()

        @app.get("/")
        def index(is_htmx: bool = Depends(air.is_htmx_request)):
            return air.H1(f"Is HTMX request: {is_htmx}")
    """
    return hx_request is not None and hx_request.lower() == "true"



from typing import Annotated

from fastapi import Depends, Request
from fastapi.testclient import TestClient
from pydantic import BaseModel, Field

import air


def test_form_sync_check():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    cheese = CheeseForm()
    cheese.validate({"name": "Parmesan", "age": "Hello"})
    assert cheese.is_valid is False
    assert cheese.errors == [
        {
            "type": "int_parsing",
            "loc": ("age",),
            "msg": "Input should be a valid integer, unable to parse string as an integer",
            "input": "Hello",
            "url": "https://errors.pydantic.dev/2.11/v/int_parsing",
        }
    ]


def test_form_validation_dependency_injection():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    app = air.Air()

    @app.post("/cheese")
    async def cheese_form(
        cheese: Annotated[CheeseForm, Depends(CheeseForm.from_request)],
    ):
        if cheese.is_valid:
            return air.Html(air.H1(cheese.data.name))  # type: ignore [union-attr]
        return air.Html(air.H1(air.RawHTML(str(len(cheese.errors)))))  # type: ignore [arg-type]

    client = TestClient(app)

    # Test with valid form data
    response = client.post("/cheese", data={"name": "cheddar", "age": 5})
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert response.text == "<!doctype html><html><h1>cheddar</h1></html>"

    # Test with invalid form data
    response = client.post("/cheese", data={})
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert response.text == "<!doctype html><html><h1>2</h1></html>"


def test_form_validation_in_view():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    app = air.Air()

    @app.post("/cheese")
    async def cheese_form(request: Request):
        cheese = await CheeseForm.from_request(request)
        if cheese.is_valid:
            return air.Html(air.H1(cheese.data.name))
        return air.Html(air.H1(air.RawHTML(str(len(cheese.errors)))))  # type: ignore [arg-type]

    client = TestClient(app)

    # Test with valid form data
    response = client.post("/cheese", data={"name": "cheddar", "age": 5})
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert response.text == "<!doctype html><html><h1>cheddar</h1></html>"

    # Test with invalid form data
    response = client.post("/cheese", data={})
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert response.text == "<!doctype html><html><h1>2</h1></html>"


def test_form_render():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    cheese = CheeseForm()

    assert (
        cheese.render()
        == '<fieldset><label>name<input name="name" type="text" id="name" /></label><label>age<input name="age" type="number" id="age" /></label></fieldset>'
    )


def test_form_render_with_values():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    cheese = CheeseForm(dict(name="Cheddar", age=3))

    assert (
        cheese.render()
        == '<fieldset><label>name<input name="name" type="text" id="name" value="Cheddar" /></label><label>age<input name="age" type="number" id="age" value="3" /></label></fieldset>'
    )


def test_form_render_in_view():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    app = air.Air()

    @app.post("/cheese")
    async def cheese_form(request: Request):
        cheese = CheeseForm()
        return air.Form(cheese.render())

    client = TestClient(app)

    # Test with valid form data
    response = client.post("/cheese", data={"name": "cheddar", "age": 5})
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert (
        response.text
        == '<form><fieldset><label>name<input name="name" type="text" id="name" /></label><label>age<input name="age" type="number" id="age" /></label></fieldset></form>'
    )


def test_form_render_with_errors():
    class CheeseModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        age: int  # type: ignore [annotation-unchecked]

    class CheeseForm(air.AirForm):
        model = CheeseModel

    cheese = CheeseForm()

    # render without validation
    html = cheese.render()
    assert "Please correct this error." not in html

    # render with validation
    cheese.validate({})
    html = cheese.render()

    assert "Please correct this error." in html

    assert (
        html
        == '<fieldset><label>name<input name="name" type="text" id="name" aria-invalid="true" /><small id="name-error">Please correct this error.</small></label><label>age<input name="age" type="number" id="age" aria-invalid="true" /><small id="age-error">Please correct this error.</small></label></fieldset>'
    )


def test_html_input_field_types():
    class ContactModel(BaseModel):
        name: str  # type: ignore [annotation-unchecked]
        email: str = Field(json_schema_extra={"email": True})  # type: ignore [annotation-unchecked]
        date_and_time: str = Field(json_schema_extra={"datedatetime-local": True})  # type: ignore [annotation-unchecked]

    class ContactForm(air.AirForm):
        model = ContactModel

    contact_form = ContactForm()
    html = contact_form.render()
    assert 'type="datedatetime-local"' in html
    assert 'type="email"' in html



"""A FastAPI-powered breath of fresh air in Python web development."""

__version__ = "0.13.0"

from starlette.staticfiles import StaticFiles as StaticFiles  # noqa

from . import layouts as layouts
from . import responses as responses
from . import svg as svg
from .applications import Air as Air
from .background import BackgroundTasks as BackgroundTasks
from .forms import AirForm as AirForm
from .requests import Request as Request
from .requests import is_htmx_request as is_htmx_request
from .responses import AirResponse as AirResponse
from .responses import RedirectResponse as RedirectResponse
from .responses import TagResponse as TagResponse
from .tags import (
    H1 as H1,
)
from .tags import (
    H2 as H2,
)
from .tags import (
    H3 as H3,
)
from .tags import (
    H4 as H4,
)
from .tags import (
    H5 as H5,
)
from .tags import (
    H6 as H6,
)
from .tags import (
    A as A,
)
from .tags import (
    Abbr as Abbr,
)
from .tags import (
    Address as Address,
)
from .tags import (
    Area as Area,
)
from .tags import (
    Article as Article,
)
from .tags import (
    Aside as Aside,
)
from .tags import (
    Audio as Audio,
)
from .tags import (
    B as B,
)
from .tags import (
    Base as Base,
)
from .tags import (
    Bdi as Bdi,
)
from .tags import (
    Bdo as Bdo,
)
from .tags import (
    Blockquote as Blockquote,
)
from .tags import (
    Body as Body,
)
from .tags import (
    Br as Br,
)
from .tags import (
    Button as Button,
)
from .tags import (
    Canvas as Canvas,
)
from .tags import (
    Caption as Caption,
)
from .tags import (
    Cite as Cite,
)
from .tags import (
    Code as Code,
)
from .tags import (
    Col as Col,
)
from .tags import (
    Colgroup as Colgroup,
)
from .tags import (
    Data as Data,
)
from .tags import (
    Datalist as Datalist,
)
from .tags import (
    Dd as Dd,
)
from .tags import (
    Del as Del,
)
from .tags import (
    Details as Details,
)
from .tags import (
    Dfn as Dfn,
)
from .tags import (
    Dialog as Dialog,
)
from .tags import (
    Div as Div,
)
from .tags import (
    Dl as Dl,
)
from .tags import (
    Dt as Dt,
)
from .tags import (
    Em as Em,
)
from .tags import (
    Embed as Embed,
)
from .tags import (
    Fieldset as Fieldset,
)
from .tags import (
    Figcaption as Figcaption,
)
from .tags import (
    Figure as Figure,
)
from .tags import (
    Footer as Footer,
)
from .tags import (
    Form as Form,
)
from .tags import (
    Head as Head,
)
from .tags import (
    Header as Header,
)
from .tags import (
    Hgroup as Hgroup,
)
from .tags import (
    Hr as Hr,
)
from .tags import (
    Html as Html,
)
from .tags import (
    I as I,
)
from .tags import (
    Iframe as Iframe,
)
from .tags import (
    Img as Img,
)
from .tags import (
    Input as Input,
)
from .tags import (
    Ins as Ins,
)
from .tags import (
    Kbd as Kbd,
)
from .tags import (
    Label as Label,
)
from .tags import (
    Legend as Legend,
)
from .tags import (
    Li as Li,
)
from .tags import (
    Link as Link,
)
from .tags import (
    Main as Main,
)
from .tags import (
    Map as Map,
)
from .tags import (
    Mark as Mark,
)
from .tags import (
    Menu as Menu,
)
from .tags import (
    Meta as Meta,
)
from .tags import (
    Meter as Meter,
)
from .tags import (
    Nav as Nav,
)
from .tags import (
    Noscript as Noscript,
)
from .tags import (
    Object as Object,
)
from .tags import (
    Ol as Ol,
)
from .tags import (
    Optgroup as Optgroup,
)
from .tags import (
    Option as Option,
)
from .tags import (
    Output as Output,
)
from .tags import (
    P as P,
)
from .tags import (
    Param as Param,
)
from .tags import (
    Picture as Picture,
)
from .tags import (
    Pre as Pre,
)
from .tags import (
    Progress as Progress,
)
from .tags import (
    Q as Q,
)
from .tags import (
    RawHTML as RawHTML,
)
from .tags import (
    Rp as Rp,
)
from .tags import (
    Rt as Rt,
)
from .tags import (
    Ruby as Ruby,
)
from .tags import (
    S as S,
)
from .tags import (
    Samp as Samp,
)
from .tags import (
    Script as Script,
)
from .tags import (
    Search as Search,
)
from .tags import (
    Section as Section,
)
from .tags import (
    Select as Select,
)
from .tags import (
    Small as Small,
)
from .tags import (
    Source as Source,
)
from .tags import (
    Span as Span,
)
from .tags import (
    Strong as Strong,
)
from .tags import (
    Style as Style,
)
from .tags import (
    Sub as Sub,
)
from .tags import (
    Summary as Summary,
)
from .tags import (
    Sup as Sup,
)
from .tags import (
    Table as Table,
)
from .tags import Tag as Tag
from .tags import (
    Tbody as Tbody,
)
from .tags import (
    Td as Td,
)
from .tags import (
    Template as Template,
)
from .tags import (
    Textarea as Textarea,
)
from .tags import (
    Tfoot as Tfoot,
)
from .tags import (
    Th as Th,
)
from .tags import (
    Thead as Thead,
)
from .tags import (
    Time as Time,
)
from .tags import (
    Title as Title,
)
from .tags import (
    Tr as Tr,
)
from .tags import (
    Track as Track,
)
from .tags import (
    U as U,
)
from .tags import (
    Ul as Ul,
)
from .tags import (
    Var as Var,
)
from .tags import (
    Video as Video,
)
from .tags import (
    Wbr as Wbr,
)
from .tags import html_to_airtags as html_to_airtags
from .templates import Jinja2Renderer as Jinja2Renderer



from fastapi.background import BackgroundTasks as FastAPIBackgroundTasks
from typing_extensions import Annotated, Any, Callable, Doc, ParamSpec

P = ParamSpec("P")


class BackgroundTasks(FastAPIBackgroundTasks):
    """
    A collection of background tasks that will be called after a response has been
    sent to the client.

    ## Example

    ```python
    from air import BackgroundTasks, Air
    from air.layouts import picocss

    app = Air()


    def write_notification(email: str, message=""):
        with open("log.txt", mode="w") as email_file:
            content = f"notification for {email}: {message}"
            email_file.write(content)


    @app.post("/send-notification/{email}")
    async def send_notification(email: str, background_tasks: BackgroundTasks):
        message = "some notification"
        background_tasks.add_task(write_notification, email, message=message)
        content = f"notification for {email}: {message}"
        return picocss(
            air.Title(content),
            air.H1(content)
        )
    ```
    """

    def add_task(
        self,
        func: Annotated[
            Callable[P, Any],
            Doc(
                """
                The function to call after the response is sent.

                It can be a regular `def` function or an `async def` function.
                """
            ),
        ],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """
        Add a function to be called in the background after the response is sent.
        """
        return super().add_task(func, *args, **kwargs)
